"""

metta Terraform

metta contrib functionality for terraform.  Primarily a terraform provisioner
plugin.

"""
from typing import Any

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT

from mirantis.testing.metta.plugin import Factory, Type
from mirantis.testing.metta.environment import Environment

from .cli import TerraformCliPlugin
from .provisioner import TerraformProvisionerPlugin, TERRAFORM_PROVISIONER_CONFIG_LABEL, TERRAFORM_VALIDATE_JSONSCHEMA

METTA_TERRAFORM_PROVISIONER_PLUGIN_ID = 'metta_terraform'
""" Terraform provisioner plugin id """


@Factory(type=Type.PROVISIONER, plugin_id=METTA_TERRAFORM_PROVISIONER_PLUGIN_ID)
def metta_plugin_factory_provisioner_terraform(
        environment: Environment, instance_id: str = "", label: str = TERRAFORM_PROVISIONER_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
    """ create an metta provisionersss dict plugin """
    return TerraformProvisionerPlugin(environment, instance_id, label, base)


METTA_TERRAFORM_CLI_PLUGIN_ID = 'metta_terraform'
""" cli plugin_id for the info plugin """


@Factory(type=Type.CLI, plugin_id=METTA_TERRAFORM_CLI_PLUGIN_ID)
def metta_terraform_factory_cli_terraform(
        environment: Environment, instance_id: str = ''):
    """ create an info cli plugin """
    return TerraformCliPlugin(environment, instance_id)


""" SetupTools EntryPoint METTA BootStrapping """


def bootstrap(environment: Environment):
    """ METTA_Terraform bootstrap

    Currently we only use this to import plugins.

    Parameters:
    -----------

    env (Environment) : an environment which should have validation config added
        to.

    """
    pass



import logging
from typing import Dict, Any

import json

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase
from mirantis.testing.metta_cli.provisioner import ProvisionerGroup

logger = logging.getLogger('metta.cli.ansible')


class AnsibleCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands """
        if self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id='metta_ansible', exception_if_missing=False) is not None:
            return {
                'contrib': {
                    'ansible': AnsibleGroup(self.environment)
                }
            }
        else:
            return {}


class AnsibleGroup():

    def __init__(self, environment: Environment):
        self.environment = environment

    def _select_provisioner(self, instance_id: str = ''):
        """ Pick a matching provisioner """
        if instance_id:
            return self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id='metta_ansible', instance_id=instance_id)
        else:
            # Get the highest priority provisioner
            return self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, plugin_id='metta_ansible')

    def info(self, provisioner: str = ''):
        """ get info about a provisioner plugin """
        fixture = self._select_provisioner(instance_id=provisioner)
        plugin = fixture.plugin

        info = {
            'fixture': {
                'type': fixture.type.value,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority
            }
        }
        info.update(plugin.info())

        return json.dumps(info, indent=2)

    def prepare(self, provisioner: str = ''):
        """ Run provisioner prepare """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.prepare()

    def apply(self, provisioner: str = ''):
        """ Run provisioner apply """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.apply()

    def destroy(self, provisioner: str = ''):
        """ Run provisioner destroy """
        provisioner = self._select_provisioner(instance_id=provisioner).plugin
        provisioner.destroy()



"""

Launchpad metta provisioner plugin

"""

import json
import yaml
import os.path
import logging
from typing import List, Dict, Any

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL
from configerus.validator import ValidationError

from mirantis.testing.metta.plugin import METTAPlugin, Type
from mirantis.testing.metta.fixtures import Fixtures, UCCTFixturesPlugin
from mirantis.testing.metta.provisioner import ProvisionerBase
from mirantis.testing.metta_docker import METTA_PLUGIN_ID_DOCKER_CLIENT
from mirantis.testing.metta_kubernetes import METTA_PLUGIN_ID_KUBERNETES_CLIENT
from mirantis.testing.metta_mirantis import METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID, METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID

from .launchpad import LaunchpadClient, METTA_USER_LAUNCHPAD_CLUSTER_PATH, METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT
from .exec_client import METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID

logger = logging.getLogger('mirantis.testing.metta.provisioner:launchpad')

METTA_LAUNCHPAD_CONFIG_LABEL = 'launchpad'
""" Launchpad config label for configuration """
METTA_LAUNCHPAD_CONFIG_ROOT_PATH_KEY = 'root.path'
""" config key for a base file path that should be used for any relative paths """
METTA_LAUNCHPAD_CONFIG_KEY = 'config'
""" which config key will provide the launchpad yml """
METTA_LAUNCHPAD_CONFIG_API_ACCESSPOINT_KEY = 'api.accesspoint'
""" config key for the API endpoint, usually the manager load-balancer """
METTA_LAUNCHPAD_CONFIG_API_USERNAME_KEY = 'api.username'
""" config key for the API username """
METTA_LAUNCHPAD_CONFIG_API_PASSWORD_KEY = 'api.password'
""" config key for the API password """
METTA_LAUNCHPAD_CONFIG_HOSTS_KEY = 'config.spec.hosts'
""" config key for the list of hosts as per the launchpad spec """
METTA_LAUNCHPAD_CLI_CONFIG_FILE_KEY = 'config_file'
""" Launchpad config cli key to tell us where to put the launchpad yml file """
METTA_LAUNCHPAD_CLI_WORKING_DIR_KEY = 'working_dir'
""" Launchpad config cli configuration working dir key """
METTA_LAUNCHPAD_CLI_CLUSTEROVERRIDE_KEY = 'cluster_name'
""" If provided, this config key will override a cluster name pulled from yaml"""
METTA_LAUNCHPAD_CLI_ACCEPTLICENSE_KEY = 'cli.accept-license'
""" If provided, this config key will tell the cli to accept the license"""
METTA_LAUNCHPAD_CLI_DISABLETELEMETRY_KEY = 'cli.disable-telemetry'
""" If provided, this config key will tell the cli to disable telemetry"""
METTA_LAUNCHPAD_CLI_DISABLETELEMETRYUPGRADECHECK_KEY = 'cli.disable-upgrade-check'
""" If provided, this config key will tell the cli to disable upgrade checks"""
METTA_LAUNCHPAD_CLI_CONFIG_DOCKER_VERSION_DEFAULT = '1.40'
""" Default value for the docker client version number.  It would be best to discover or config this."""

METTA_LAUNCHPAD_VALIDATE_JSONSCHEMA = {
    'type': 'object',
    'properties': {
        'config': {
            'type': ['object', 'null'],
            'properties': {
                'apiVersion': {'type': 'string'}
            },
        },
        'api': {
            'type': 'object',
            'properties': {
                'endpoint': {'type': ['string', 'null']}
            },
        },
        'cli': {
            'accept-license': {'type': 'bool'},
            'disable-telemetry': {'type': 'bool'},
        },
        'cluster_name': {
            'type': 'string'
        },
        'config_file': {
            'type': 'string'
        },
        'working_dir': {
            'type': 'string'
        }
    },
    'required': ['config_file', 'config', 'api']
}
""" Validation jsonschema for terraform config contents """
METTA_LAUNCHPAD_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: METTA_LAUNCHPAD_VALIDATE_JSONSCHEMA
}
""" configerus validation target to matche the above config, which relates to the bootstrap in __init__.py """


class LaunchpadProvisionerPlugin(ProvisionerBase, UCCTFixturesPlugin):
    """ Launchpad provisioner class

    Provision a system using Mirantis launchpad

    """

    def __init__(self, environment, instance_id,
                 label: str = METTA_LAUNCHPAD_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
        """ Run the super constructor but also set class properties """
        ProvisionerBase.__init__(self, environment, instance_id)
        UCCTFixturesPlugin.__init__(self)

        """ Set an empty client, populated after the backend is provisioned """
        self.client = None
        """ A configured LaunchpadClient """

        self.downloaded_bundle_users = []
        """ track user bundles that have been downloaded to avoid unecessary repeats """

        self.config_label = label
        self.config_base = base

        """ load all of the launchpad configuration """
        launchpad_config = self.environment.config.load(label)

        # Run confgerus validation on the config using our above defined
        # jsonschema
        try:
            launchpad_config.get(
                base, validator=METTA_LAUNCHPAD_VALIDATE_TARGET)
        except ValidationError as e:
            raise ValueError(
                "Launchpad config failed validation: {}".format(e)) from e

        self.working_dir = launchpad_config.get(
            [base, METTA_LAUNCHPAD_CLI_WORKING_DIR_KEY])
        """ if launchpad needs to be run in a certain path, set it with this config """
        if not self.working_dir:
            self.working_dir = METTA_LAUNCHPADCLIENT_WORKING_DIR_DEFAULT
        if not os.path.isabs(self.working_dir):
            # did a relative path root get passed in as config?
            root_path = self.backend_output_name = launchpad_config.get(
                [base, METTA_LAUNCHPAD_CONFIG_ROOT_PATH_KEY])
            if root_path:
                self.working_dir = os.path.join(root_path, self.working_dir)
            self.working_dir = os.path.abspath(self.working_dir)

        """ Retrieve the configuration the client plugin """
        try:
            self.launchpad_config = launchpad_config.get(
                [base, METTA_LAUNCHPAD_CONFIG_KEY])
            """ config source of launchpad yaml """
        except KeyError as e:
            raise ValueError(
                "Could not find launchpad configuration.")

        # decide on a path for the runtime launchpad.yml file
        self.config_file = launchpad_config.get(
            [base, METTA_LAUNCHPAD_CLI_CONFIG_FILE_KEY], default=METTA_LAUNCHPAD_CLI_CONFIG_FILE_DEFAULT)
        if not os.path.isabs(self.config_file):
            # A relative [ath for the config file is expected to be relative to
            # the working dir]
            self.config_file = os.path.abspath(
                os.path.join(self.working_dir, self.config_file))

        cluster_name_override = launchpad_config.get(
            [base, METTA_LAUNCHPAD_CLI_CLUSTEROVERRIDE_KEY], default='')
        """ Can hardcode the cluster name if it can't be take from the yaml file """

        accept_license = launchpad_config.get(
            [base, METTA_LAUNCHPAD_CLI_ACCEPTLICENSE_KEY], default=True)
        """ should the client accept the license """
        disable_telemetry = launchpad_config.get(
            [base, METTA_LAUNCHPAD_CLI_DISABLETELEMETRY_KEY], default=True)
        """ should the client disable telemetry """
        disable_upgrade_check = launchpad_config.get(
            [base, METTA_LAUNCHPAD_CLI_DISABLETELEMETRYUPGRADECHECK_KEY], default=True)
        """ should the client disable upgrade checks """

        logger.debug("Creating Launchpad API client")
        self.client = LaunchpadClient(
            config_file=self.config_file,
            working_dir=self.working_dir,
            cluster_name_override=cluster_name_override,
            accept_license=accept_license,
            disable_telemetry=disable_telemetry,
            disable_upgrade_check=disable_upgrade_check)

        # If we can, it makes sense to build the docker and k8s client fixtures now.
        # This will only be possible in cases where we have an existing config file
        # and we can download a client bundle.
        # We try that here, even though it is verbose and ugly, so that we have
        # the clients available for introspection.
        # We probably shouldn't, but it allows some flexibility.
        if os.path.exists(self.config_file):
            try:
                self._make_fixtures()
            except BaseException:
                logger.debug(
                    "Launchpad couldn't initialize some plugins as we don't have enough information to go on.")
                # we most likely failed because we don't have enough info get
                # make fixtures from

    def info(self, deep: bool = False):
        """ get info about a provisioner plugin """
        plugin = self
        client = self.client

        info = {
            'plugin': {
                'config_label': plugin.config_label,
                'config_base': plugin.config_base,
                'downloaded_bundle_users': plugin.downloaded_bundle_users,
                'working_dir': plugin.working_dir,
            },
            'client': {
                'cluster_name_override': client.cluster_name_override,
                'config_file': client.config_file,
                'working_dir': client.working_dir,
                'bin': client.bin
            },
            'config': self.launchpad_config,
        }

        if deep:
            try:
                info['config'] = client.describe_config()
                info['bundles'] = {user: client.bundle(
                    user) for user in client.bundle_users()}
            except Exception:
                pass

            fixtures = {}
            for fixture in self.get_fixtures().to_list():
                fixture_info = {
                    'fixture': {
                        'type': fixture.type.value,
                        'plugin_id': fixture.plugin_id,
                        'instance_id': fixture.instance_id,
                        'priority': fixture.priority,
                    }
                }
                if hasattr(fixture.plugin, 'info'):
                    plugin_info = fixture.plugin.info()
                    if isinstance(plugin_info, dict):
                        fixture_info.update(plugin_info)
                fixtures[fixture.instance_id] = fixture_info
            info['fixtures'] = fixtures

        info['helper'] = {
            'commands': {
                'apply': "{bin} apply -c {config_file}".format(workingpathcd=("cd {} && ".format(client.working_dir) if not client.working_dir == '.' else ''), bin=client.bin, config_file=client.config_file),
                'client-config': "{bin} client-config -c {config_file} {user}".format(workingpathcd=("cd {} && ".format(client.working_dir) if not client.working_dir == '.' else ''), bin=client.bin, config_file=client.config_file, user='admin')
            }
        }

        return info

    def prepare(self):
        """ Prepare the provisioning cluster for install

        We ignore this.

        """
        logger.info(
            'Running Launchpad Prepare().  Launchpad has no prepare stage.')

    def apply(self):
        """ bring a cluster up

        We assume that the cluster is running and the we can pull the required
        yaml from an output fixture in the environment.

        This plugin needs an output fixture, probably of dict type.  It will
        Pull that structure for the launchpad yaml config file and dump it into
        its config path.
        The provisioner can find an output directly from the environment, or
        from a specific fixture source.  If you want the output to come from
        only a specific backend fixture then make sure that a "backend" config
        exists, otherwise just use an "output" config.

        Raises:
        -------

        ValueError if the object has been configured (prepare) with config that
            doesn't work, or if the backend doesn't give valid yml

        Exception if launchpad fails.

        """

        try:
            logger.info(
                "Using launchpad to install products onto backend cluster")
            self._write_launchpad_file()
            self.client.apply()
        except Exception as e:
            raise Exception("Launchpad failed to install") from e

        # Rebuild the fixture list now that we have installed
        self._make_fixtures(reload=True)

    def destroy(self):
        """ We don't bother uninstalling at this time

        we do:
        1. delete the generated config file
        2. ask the client to remove any downloaded client bundles

        """
        if os.path.exists(self.config_file):
            os.remove(self.config_file)
        self.client.rm_client_bundles()
        self.client.reset()

    """ CLUSTER INTERACTION """

    def _write_launchpad_file(self):
        """ write the config state to the launchpad file """
        try:
            """ load all of the launchpad configuration, force a reload to get up to date contents """
            launchpad_config = self.environment.config.load(
                self.config_label, force_reload=True)
            self.launchpad_config = launchpad_config.get(
                [self.config_base, METTA_LAUNCHPAD_CONFIG_KEY])
            """ config source of launchpad yaml """
        except KeyError as e:
            raise ValueError(
                "Could not find launchpad configuration from config.")

        # write the launchpad output to our yaml file target (after creating
        # the path)
        os.makedirs(
            os.path.dirname(
                os.path.realpath(
                    self.config_file)),
            exist_ok=True)
        with open(os.path.realpath(self.config_file), 'w') as file:
            yaml.dump(self.launchpad_config, file)

    def _make_fixtures(self, user: str = 'admin',
                       reload: bool = False) -> Fixtures:
        """ Build fixtures for all of the clients """

        launchpad_config = self.environment.config.load(
            self.config_label, force_reload=reload)
        """ get fresh values for the launchpad config (in case it has changed) """

        bundle_info = self._mke_client_bundle(user, reload)
        """ holds retrieved bundle information from MKE for all client configs. """

        # KUBE Client

        kube_config = os.path.join(bundle_info['path'], 'kube.yml')
        if not os.path.exists(kube_config):
            raise NotImplemented(
                "Launchpad was asked for a kubernetes client, but not kube config file was in the client bundle.  Are you sure this is a kube cluster?")

        instance_id = "launchpad-{}-{}-{}-client".format(
            self.instance_id, METTA_PLUGIN_ID_KUBERNETES_CLIENT, user)
        fixture = self.environment.add_fixture(
            type=Type.CLIENT,
            plugin_id=METTA_PLUGIN_ID_KUBERNETES_CLIENT,
            instance_id=instance_id,
            priority=70,
            arguments={'kube_config_file': kube_config})
        self.fixtures.add_fixture(fixture)

        # DOCKER CLIENT
        #
        # @NOTE we pass in a docker API constraint because I ran into a case where the
        #   client failed because the python library was ahead in API version

        try:
            host = bundle_info['Endpoints']['docker']['Host']
            cert_path = bundle_info['tls_paths']['docker']
        except TypeError as e:
            logger.error(
                "Could not read client bundle properly: %s",
                bundle_info['Endpoints']['docker']['Host'])
            raise e

        instance_id = "launchpad-{}-{}-{}-client".format(
            self.instance_id, METTA_PLUGIN_ID_DOCKER_CLIENT, user)
        fixture = self.environment.add_fixture(
            type=Type.CLIENT,
            plugin_id=METTA_PLUGIN_ID_DOCKER_CLIENT,
            instance_id=instance_id,
            priority=70,
            arguments={'host': host, 'cert_path': cert_path, 'version': METTA_LAUNCHPAD_CLI_CONFIG_DOCKER_VERSION_DEFAULT})
        self.fixtures.add_fixture(fixture)

        # EXEC CLIENT
        #
        instance_id = "launchpad-{}-{}-{}-client".format(
            self.instance_id, METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID, user)
        fixture = self.environment.add_fixture(
            type=Type.CLIENT,
            plugin_id=METTA_LAUNCHPAD_EXEC_CLIENT_PLUGIN_ID,
            instance_id=instance_id,
            priority=70,
            arguments={'client': self.client})
        self.fixtures.add_fixture(fixture)

        # The following clients both need the LB, username and password

        api_accesspoint = self.launchpad_config = launchpad_config.get(
            [self.config_base, METTA_LAUNCHPAD_CONFIG_API_ACCESSPOINT_KEY])
        api_username = self.launchpad_config = launchpad_config.get(
            [self.config_base, METTA_LAUNCHPAD_CONFIG_API_USERNAME_KEY])
        api_password = self.launchpad_config = launchpad_config.get(
            [self.config_base, METTA_LAUNCHPAD_CONFIG_API_PASSWORD_KEY])

        hosts = self.launchpad_config = launchpad_config.get(
            [self.config_base, METTA_LAUNCHPAD_CONFIG_HOSTS_KEY])
        if hosts is None:
            hosts = []

        # MKE Client
        #
        instance_id = "launchpad-{}-{}-{}-client".format(
            self.instance_id, METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID, user)
        mke_hosts = [host for host in hosts if host['role'] in ['manager']]
        if len(mke_hosts) > 0:
            fixture = self.environment.add_fixture(
                type=Type.CLIENT,
                plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID,
                instance_id=instance_id,
                priority=70,
                arguments={'accesspoint': api_accesspoint, 'username': api_username, 'password': api_password, 'hosts': mke_hosts})
            self.fixtures.add_fixture(fixture)

        # MSR Client
        #
        instance_id = "launchpad-{}-{}-{}-client".format(
            self.instance_id, METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID, user)
        msr_hosts = [host for host in hosts if host['role'] in ['msr']]
        if len(msr_hosts) > 0:
            fixture = self.environment.add_fixture(
                type=Type.CLIENT,
                plugin_id=METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID,
                instance_id=instance_id,
                priority=70,
                arguments={'accesspoint': None, 'username': api_username, 'password': api_password, 'hosts': msr_hosts})
            self.fixtures.add_fixture(fixture)

    def _mke_client_bundle(self, user: str, reload: bool = False):
        """ Retrieve the MKE Client bundle metadata using the client """
        assert self.client, "Don't have a launchpad client configured yet"
        return self.client.bundle(user, reload)



"""

Dummy workload plugin

"""
from typing import Dict, Any
import logging

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.workload import WorkloadBase
from mirantis.testing.metta.fixtures import UCCTFixturesPlugin

logger = logging.getLogger('metta.contrib.dummy.workload')


class DummyWorkloadPlugin(WorkloadBase, UCCTFixturesPlugin):
    """ Dummy workload class """

    def __init__(self, environment: Environment, instance_id: str,
                 fixtures: Dict[str, Dict[str, Any]] = {}):
        """ Run the super constructor but also set class properties

        Parameters:
        -----------

        outputs (Dict[Dict]) : pass in a dictionary which defines outputs that
            should be returned

        clients (Dict[Dict]) : pass in a dictionary which defines which clients
            should be requested when working on a provisioner

        """
        WorkloadBase.__init__(self, environment, instance_id)

        fixtures = environment.add_fixtures_from_dict(plugin_list=fixtures)
        """ All fixtures added to this dummy plugin. """
        UCCTFixturesPlugin.__init__(self, fixtures)



import logging
from typing import Dict, Any, List

import json

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase

logger = logging.getLogger('metta.cli.output')


class OutputCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands """
        return {
            'output': OutputGroup(self.environment)
        }


class OutputGroup():

    def __init__(self, environment: Environment):
        self.environment = environment

    def list(self):
        """ List all outputs """
        list = [
            plugin.instance_id for plugin in self.environment.fixtures.fixtures(
                type=Type.OUTPUT)]

        json.dumps(list)

    def info(self, plugin_id: str = '',
             instance_id: str = '', deep: bool = False):
        """ List all outputs """
        details = []
        for plugin in self.environment.fixtures.get_plugins(type=Type.OUTPUT):
            info = {
                'fixture': {
                    'type': fixture.type.value,
                    'plugin_id': fixture.plugin_id,
                    'instance_id': fixture.instance_id,
                    'priority': fixture.priority,
                }
            }

            if deep and hasattr(fixture.plugin, 'info'):
                plugin_info = fixture.plugin.info()
                if isinstance(plugin_info, dict):
                    info.update(plugin_info)

            list.append(info)

        json.dumps(list)



"""

Common METTA plugins

"""
from typing import Dict, Any

from configerus import Config
from configerus.loaded import LOADED_KEY_ROOT
from configerus.plugin import FormatFactory

from mirantis.testing.metta.plugin import Factory, Type
from mirantis.testing.metta.environment import Environment

from .common_config import add_common_config
from .dict_output import DictOutputPlugin
from .text_output import TextOutputPlugin
from .combo_provisioner import ComboProvisionerPlugin, COMBO_PROVISIONER_CONFIG_LABEL
from .binhelper_utility import DownloadableExecutableUtility, METTA_PLUGIN_ID_UTILITY_BINHELPER, BINHELPER_UTILITY_CONFIG_LABEL
from .config_format_output import ConfigFormatOutputPlugin

METTA_PLUGIN_ID_OUTPUT_DICT = 'dict'
""" output plugin_id for the dict plugin """


@Factory(type=Type.OUTPUT, plugin_id=METTA_PLUGIN_ID_OUTPUT_DICT)
def metta_plugin_factory_output_dict(
        environment: Environment, instance_id: str = '', data: Dict = {}, validator: str = ''):
    """ create an output dict plugin """
    return DictOutputPlugin(environment, instance_id, data, validator)


METTA_PLUGIN_ID_OUTPUT_TEXT = 'text'
""" output plugin_id for the text plugin """


@Factory(type=Type.OUTPUT, plugin_id=METTA_PLUGIN_ID_OUTPUT_TEXT)
def metta_plugin_factory_output_text(
        environment: Environment, instance_id: str = '', text: str = ''):
    """ create an output text plugin """
    return TextOutputPlugin(environment, instance_id, text)


METTA_PLUGIN_ID_PROVISIONER_COMBO = 'combo'
""" provisioner plugin_id for the combo plugin """


@Factory(type=Type.PROVISIONER, plugin_id=METTA_PLUGIN_ID_PROVISIONER_COMBO)
def metta_plugin_factory_provisioner_combo(
        environment: Environment, instance_id: str = '', label: str = COMBO_PROVISIONER_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
    """ create a provisioner combo plugin """
    return ComboProvisionerPlugin(
        environment, instance_id, label=label, base=base)


@Factory(type=Type.UTILITY, plugin_id=METTA_PLUGIN_ID_UTILITY_BINHELPER)
def metta_plugin_factory_utility_binhelper(
        environment: Environment, instance_id: str = '', label: str = BINHELPER_UTILITY_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
    """ create a bin-helper utility plugin """
    return DownloadableExecutableUtility(
        environment, instance_id, label=label, base=base)


""" Congiferus formatter for metta output plugins """

PLUGIN_ID_FORMAT_OUTPUT = 'output'
""" Format plugin_id for the configerus output format plugin """


@FormatFactory(plugin_id=PLUGIN_ID_FORMAT_OUTPUT)
def plugin_factory_format_output(config: Config, instance_id: str = ''):
    """ create an format plugin which replaces from output contents """
    return ConfigFormatOutputPlugin(config, instance_id)


""" METTA bootstraps that we will use on config objects """


def bootstrap(environment: Environment):
    """ METTA Bootstrapper - don't actually do anything """
    pass

# @TODO this should be renamed on the next major version bump. 
def bootstrap_common(environment: Environment):
    """ metta configerus bootstrap

    Add some common Mirantis specific config options

    Add some common configerus sources for common data and common config source
    paths. Some of the added config is dynamic interpretation of environment,
    while also some default config paths are added if they can be interpeted

    @see .config.add_common_config() for details

    """
    add_common_config(environment)

    # Add a configerus output formatter which can interpret environemnt
    # outputs.
    output_formatter = environment.config.add_formatter(
        plugin_id=PLUGIN_ID_FORMAT_OUTPUT,
        instance_id=PLUGIN_ID_FORMAT_OUTPUT,
        priority=40)
    output_formatter.set_environemnt(environment)



"""

DUMMY Plugin testing.

As toolbox operations are quite context sensitive, and could involve creating
infrastructure, we don't have unit testing for them

Here we test the overall toolbox workflows using dummy plugins.

"""
import logging
import unittest

# Configerus imports needed to create a config object
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT

# METTA components we will need to access the toolbox
from mirantis.testing.metta import new_environment, environment_names, get_environment
from mirantis.testing.metta.plugin import Type
# Used for type hinting
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixtures import Fixtures

# imports used only for type testing
from mirantis.testing.metta_dummy.provisioner import DummyProvisionerPlugin
from mirantis.testing.metta_dummy.client import DummyClientPlugin
from mirantis.testing.metta_dummy.workload import DummyWorkloadPlugin

logger = logging.getLogger("test_dummy")
logger.setLevel(logging.INFO)


""" Config for generating dummy plugins

 What is this thing

Well it is a fixtures list which will get passed to add_fixtures_from_config()
on the Environment object.
Plugins are in linear order key=>plugin_definition.

Some plugins/fixtures take 'arguments' which are passed to their constructors.
The Dummy plugins tend to take a 'fixtures' argument which they use to build
more plugins.  This is why it has a deep nested look.

WHAT WEIRD THINGS HAVE WE DONE:

1. plugin type:  sometimes we use the actual Type.XXXX.value, sometimes a short
        string, and a long string. We need to make sure that all of them work.

"""

CONFIG_DATA = {
    'jsonschema': {
        'plugin': {
            "type": "object",
            "properties": {
                "type": {"type": "string"},
                "plugin_id": {"type": "string"},
            }
        }
    },
    'fixtures': {
        'prov1': {
            'type': 'PROVISIONER',
            'plugin_id': 'dummy',
            'arguments': {
                # Dummy provisioner fixtures
                'fixtures': {
                    # 1. A dummy client fixture that holds 2 dummy outputs
                    'client1': {
                        'type': 'client',
                        'plugin_id': 'dummy',
                        'arguments': {
                            'fixtures': {
                                # 1.a. First client output is a dummy text
                                'output1': {
                                    'type': 'output',
                                    'plugin_id': 'text',
                                    'arguments': {
                                        # The text output takes a string 'data'
                                        # constructor argument
                                        'text': "prov client one output one"
                                    }
                                },
                                # 1.b. Second client output is a dummy dict
                                'output2': {
                                    'type': 'metta.plugin.output',
                                    'plugin_id': 'dict',
                                    'arguments': {
                                        # The dict output takes a dict 'data'
                                        # constructor argument
                                        'data': {
                                            '1': {
                                                '1': "prov client one output two data one.one"
                                            }
                                        }
                                    }
                                }
                            }
                        }
                    },
                    # 2. Dummy text output
                    'output1': {
                        'type': Type.OUTPUT.value,
                        'plugin_id': 'text',
                        'arguments': {
                            'text': "prov dummy output one"
                        }
                    }
                }
            }
        },
        'work1': {
            'type': Type.WORKLOAD.value,
            'plugin_id': 'dummy',
        },
        'work2': {
            'type': 'workload',
            'plugin_id': 'dummy',
            'arguments': {
                'fixtures': {
                    'output1': {
                        'type': 'output',
                        'plugin_id': 'text',
                        'arguments': {
                            # The text output takes a string 'data' constructor
                            # argument
                            'text': "workload two dummy output one"
                        }
                    },
                    'client1': {
                        'type': 'client',
                        'plugin_id': 'dummy'
                    }
                }
            }
        }
    }
}

""" TESTS """


class DummyTesting(unittest.TestCase):

    def _dummy_environment(self) -> Environment:
        """ Create an environment object, add the common config source data """
        if not 'dummy' in environment_names():
            environment = new_environment(
                name='dummy', additional_metta_bootstraps=['metta_dummy'])
            environment.config.add_source(
                PLUGIN_ID_SOURCE_DICT).set_data(CONFIG_DATA)

            # this looks magical, but it works because we have structured the
            # fixtures DICT to match default values
            environment.add_fixtures_from_config()

        return get_environment(name='dummy')

    def _dummy_provisioner(self) -> Fixtures:
        """ Return a Fixtures of all WORKLOAD plugins """
        environment = self._dummy_environment()
        return environment.fixtures.get_plugin(type=Type.PROVISIONER)

    def _dummy_workloads(self) -> Fixtures:
        """ Return a Fixtures of all WORKLOAD plugins """
        environment = self._dummy_environment()
        return environment.fixtures.get_filtered(type=Type.WORKLOAD)

    """ Tests """

    def test_dummy_config_basics(self):
        """ some sanity testing on the shared config object """
        environment = self._dummy_environment()

        dummy_config = environment.config.load('fixtures')
        self.assertEqual(dummy_config.get('prov1.plugin_id'), "dummy")

        dummy_config.get('work1', validator="jsonschema:plugin")

    def test_provisioner_sanity(self):
        """ some sanity testing on loading a provisioner """
        provisioner = self._dummy_provisioner()
        self.assertIsInstance(provisioner, DummyProvisionerPlugin)

    def test_workloads_sanity(self):
        """ test that we can load the workloads """
        workloads = self._dummy_workloads()

        workload_one = workloads.get_plugin(instance_id='work1')
        self.assertIsInstance(workload_one, DummyWorkloadPlugin)

        with self.assertRaises(KeyError):
            workloads['does.not.exist']

    def test_provisioner_workflow(self):
        """ test that the provisioner can follow a decent workflow """
        provisioner = self._dummy_provisioner()

        provisioner.prepare()
        provisioner.apply()

        # ...

        provisioner.destroy()

    def test_workloads_outputs(self):
        """ test that the dummy workload got its outputs from configuration """
        workloads = self._dummy_workloads()
        workload_two = workloads.get_plugin(instance_id='work2')

        self.assertEqual(workload_two.get_output(
            instance_id='output1').get_output(), "workload two dummy output one")

    def test_provisioner_outputs(self):
        """ test that the provisioner produces the needed clients """
        provisioner = self._dummy_provisioner()
        provisioner.prepare()

        # check that we can get an output from a provisioner
        provisioner_output_dummy = provisioner.get_output(
            instance_id='output1')
        self.assertEqual(
            provisioner_output_dummy.get_output(),
            "prov dummy output one")

        # make sure that an error is raised if the key doesn't exist
        with self.assertRaises(KeyError):
            provisioner.get_output(instance_id='does not exist')

    def test_provisioner_clients(self):
        """ test that the provisioner produces the needed clients """
        provisioner = self._dummy_provisioner()
        provisioner.prepare()

        # two ways to get the same client in this case
        client_one = provisioner.get_client(instance_id='client1')
        self.assertIsInstance(client_one, DummyClientPlugin)
        self.assertEqual(client_one.instance_id, 'client1')
        client_dummy = provisioner.get_client(plugin_id='dummy')
        self.assertIsInstance(client_dummy, DummyClientPlugin)
        self.assertEqual(client_dummy.instance_id, 'client1')

        # make sure that an error is raised if the key doesn't exist
        with self.assertRaises(KeyError):
            provisioner.get_client(plugin_id='does not exist')

    def test_clients(self):
        """ test that the provisioner clients behave like clients """
        provisioner = self._dummy_provisioner()
        provisioner.prepare()

        client_one = provisioner.get_client(instance_id='client1')
        self.assertIsInstance(client_one, DummyClientPlugin)

        # test that the dummy plugin can load a text output
        client_one_output = client_one.get_output(instance_id='output1')
        self.assertEqual(
            client_one_output.get_output(),
            "prov client one output one")

        # test that the dummy plugin can load a dict output
        client_two_output = client_one.get_output(instance_id='output2')
        # Test dict as a loaded config plugin
        self.assertEqual(
            client_two_output.get_output('1.1'),
            "prov client one output two data one.one")



"""

Kubernetes workload plugins

"""

import logging
from typing import List, Any

import kubernetes

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.workload import WorkloadBase, WorkloadInstanceBase

logger = logging.getLogger('metta.contrib.kubernetes.workload.deployment')

KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_LABEL = 'kubernetes'
KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_BASE = 'workload.deployment'

KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_NAMESPACE = "namespace"
KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY = "body"


class KubernetesDeploymentWorkloadPlugin(WorkloadBase):
    """ Kubernetes workload class """

    def __init__(self, environment, instance_id,
                 label: str = KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_LABEL, base: Any = KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_BASE):
        """ Run the super constructor but also set class properties

        This implements the args part of the client interface.

        Here we expect to receive a path to a KUBECONFIG file with a context set
        and we create a Kubernetes client for use.  After that this can provide
        Core api clients as per the kubernetes SDK

        Parameters:
        -----------

        config_file (str): String path to the kubernetes config file to use

        """
        WorkloadBase.__init__(self, environment, instance_id)

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

    def create_instance(self, fixtures: Fixtures):
        """ Create a workload instance from a set of fixtures

        Parameters:
        -----------

        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes client plugin.

        """

        try:
            client = fixtures.get_plugin(
                type=Type.CLIENT, plugin_id='metta_kubernetes')
        except KeyError as e:
            raise NotImplementedError(
                "Workload could not find the needed client: {}".format('metta_kubernetes'))

        workload_config = self.environment.config.load(self.config_label)

        name = workload_config.get(
            [self.config_base, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY, 'metadata.name'], default='not-declared')
        namespace = workload_config.get(
            [self.config_base, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_NAMESPACE])
        body = workload_config.get(
            [self.config_base, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY])

        return KubernetesDeploymentWorkloadInstance(
            client, namespace, name, body)

    def info(self):
        """ Return dict data about this plugin for introspection """
        workload_config = self.environment.config.load(self.config_label)

        return {
            'workload': {
                'deployment': {
                    'namespace': workload_config.get(
                        [self.config_base, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_NAMESPACE]),
                    'body': workload_config.get(
                        [self.config_base, KUBERNETES_DEPLOYMENT_WORKLOAD_CONFIG_KEY_BODY])
                },
                'required_fixtures': {
                    'kubernetes': {
                        'type': Type.CLIENT.value,
                        'plugin_id': 'metta_kubernetes'
                    }
                }
            }
        }


class KubernetesDeploymentWorkloadInstance(WorkloadInstanceBase):

    def __init__(self, client, namespace, name, body):
        self.client = client
        self.namespace = namespace
        self.name = name
        self.body = body

        self.deployment = None

        self.read()

    def read(self):
        """ retrieve the deployment job """
        if self.deployment is None:
            apps_v1 = self.client.get_api('AppsV1Api')

            try:
                self.deployment = apps_v1.read_namespaced_deployment(
                    self.name, self.namespace)
            except Exception:
                self.deployment = None

        return self.deployment

    def apply(self):
        """ Run the workload

        @NOTE Needs a kubernetes client fixture to run.  Use .set_fixtures() first

        """

        apps_v1 = self.client.get_api('AppsV1Api')
        self.deployment = apps_v1.create_namespaced_deployment(
            body=self.body, namespace=self.namespace)

        return self.deployment

    def destroy(self):
        """ destroy any created resources """

        body = kubernetes.client.V1DeleteOptions(
            propagation_policy='Foreground',
            grace_period_seconds=5)

        apps_v1 = self.client.get_api('AppsV1Api')
        status = apps_v1.delete_namespaced_deployment(
            name=self.name, namespace=self.namespace, body=body)

        self.deployment = None

        return status



"""

Mirantis MKE API Client

"""

import logging
import requests
import json
from typing import Dict, List
from enum import Enum

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.client import ClientBase
from mirantis.testing.metta.cli import CliBase


logger = logging.getLogger('metta.contrib.metta_mirantis.client.mkeapi')


METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID = 'metta_mirantis_client_mke'
""" Mirantis MKE API Client plugin id """


class MKEAPICliPlugin(CliBase):
    """ Metta CLI plugin which injects the MKE API Client CLI commands into the CLI handler """

    def fire(self):
        """ return a dict of commands for aucnhpad provisioenrs if one hase been registered."""
        if self.environment.fixtures.get_fixture(
                type=Type.CLIENT, plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID, exception_if_missing=False) is not None:
            return {
                'contrib': {
                    'mke': MKEAPICliGroup(self.environment)
                }
            }
        else:
            return {}


class MKEAPICliGroup():
    """ Metta CLI plugin which actually provides the MKE API Client CLI commands """

    def __init__(self, environment: Environment):
        self.environment = environment

    def _select_fixture(self, instance_id: str = ''):
        """ Pick a matching fixture in case there are more than one """
        if instance_id:
            return self.environment.fixtures.get_fixture(
                type=Type.CLIENT, plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID, instance_id=instance_id)
        else:
            # Get the highest priority fixture
            return self.environment.fixtures.get_fixture(
                type=Type.CLIENT, plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID)

    def info(self, instance_id: str = '', deep: bool = False):
        """ get info about a plugin """
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin

        info = {
            'fixture': {
                'type': fixture.type.value,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        info.update(plugin.info(deep))

        return json.dumps(info, indent=2)

    def version(self, instance_id: str = ''):
        """ get info about a plugin """
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_version(), indent=2)

    def ping(self, instance_id: str = '', node: int = None):
        """ check if we can ping """
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return "OK" if plugin.api_ping(node) else "FAIL"

    def pingall(self, instance_id: str = ''):
        """ check if we can ping all of the nodes directly """
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        ping = {}
        for index in range(0, plugin.host_count()):
            try:
                plugin.api_ping(index)
                ping[plugin._node_address(index)] = True
            except BaseException:
                ping[plugin._node_address(index)] = False

        return json.dumps(ping, indent=2)

    def id(self, instance_id: str = ''):
        """ cget auth id """
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_id(), indent=2)

    def nodes(self, instance_id: str = '', node_id: str = ''):
        """ list swarm nodes """
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_nodes(node_id), indent=2)

    def services(self, instance_id: str = '', service_id: str = ''):
        """ list swarm services """
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_services(service_id), indent=2)

    def tasks(self, instance_id: str = '', task_id: str = ''):
        """ list swarm tasks """
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin.api_tasks(task_id), indent=2)

    def auth(self, instance_id: str = ''):
        """ check if we can ping """
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        return json.dumps(plugin._auth_headers(), indent=2)


class MKENodeState(Enum):
    """ MKE Node state in the node status API response """
    UNKNOWN = 'unknown'
    DOWN = 'down'
    READY = 'ready'
    DISCONNECTED = 'disconnected'

    def match(self, compare: str) -> bool:
        """ allow for string comparisons """
        return self.value == compare


class MKEAPIClientPlugin(ClientBase):
    """ Metta Client plugin for API Connections to MKE """

    def __init__(self, environment: Environment, instance_id: str,
                 accesspoint: str, username: str, password: str, hosts: List[Dict]):
        """

        Parameters:
        -----------

        Standard Metta plugin parameters:

        environment (Environment) : env in which this plugin exists
            can be used to get access to config, other fixtures etc.
        instance_id (str) : string ID for this plugin to self-identify

        Plugin specific parameters:

        accesspoint (str) : API URL endpoint

        username / password (str/str) : API authentication credentials

        hosts (List[Dict]) : List of host definition dicts used to define hosts
            that can respond to API requests in the cluster.  This can be used
            to allow directly accessing the API through a specific host.
            The first node in the list is used as an API endpoint if no endpoint
            is specified.

        """
        ClientBase.__init__(self, environment, instance_id)

        self.accesspoint = accesspoint
        self.username = username
        self.password = password

        self.hosts = hosts
        """ List of hosts """

        if self.accesspoint is None:
            # use the first host as an accesspoint if none was delivered
            self.accesspoint = self._node_address(0)

        self.verify = False
        """ should we verify ssl certs """
        if not self.verify:
            requests.packages.urllib3.disable_warnings(
                requests.packages.urllib3.exceptions.InsecureRequestWarning)

        self.auth_token = None
        """ hold the bearer auth token if created by ._auth_headers() (cache) """

    def host_count(self):
        """ Return integer host count for MKE cluster """
        return len(self.hosts)

    def info(self, deep: bool = False):
        """ return information about the plugin """
        info = {
            'api': {
                'accesspoint': self.accesspoint,
                'username': self.username,
            },
            'hosts': self.hosts
        }

        if deep:
            info['id'] = self.api_id()
            info['version'] = self.api_version()
            info['api_info'] = self.api_info()

        return info

    def api_ping(self, node: int = None) -> bool:
        """ Check the API ping response """
        if node is not None:
            endpoint = self._accesspoint_url('_ping', node=node)
        else:
            endpoint = self._accesspoint_url('_ping')

        with requests.get(endpoint, verify=self.verify) as response:
            response.raise_for_status()
            return response.ok

    def api_id(self) -> Dict:
        """ retrieve the API id """
        with requests.get(self._accesspoint_url('id'), headers=self._auth_headers(), verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_version(self) -> Dict:
        """ retrieve version """
        with requests.get(self._accesspoint_url('version'), headers=self._auth_headers(), verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_info(self) -> Dict:
        """ retrieve the API info """
        with requests.get(self._accesspoint_url('info'), headers=self._auth_headers(), verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_nodes(self, node_id: str = '') -> Dict:
        """ retrieve the API nodes """
        endpoint = 'nodes/{id}'.format(id=node_id) if node_id else 'nodes'
        with requests.get(self._accesspoint_url(endpoint), headers=self._auth_headers(), verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_services(self, service_id: str = '') -> Dict:
        """ retrieve the API services """
        endpoint = 'services/{id}'.format(
            id=service_id) if service_id else 'services'
        with requests.get(self._accesspoint_url(endpoint), headers=self._auth_headers(), verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def api_tasks(self, task_id: str = '') -> Dict:
        """ retrieve the API tasks """
        endpoint = 'tasks/{id}'.format(id=task_id) if task_id else 'tasks'
        with requests.get(self._accesspoint_url(endpoint), headers=self._auth_headers(), verify=self.verify) as response:
            response.raise_for_status()
            return json.loads(response.content)

    def _auth_headers(self):
        """ get an auth token """
        if self.auth_token is None:
            data = {
                "password": self.password,
                "username": self.username
            }
            with requests.post(self._accesspoint_url('auth/login'), data=json.dumps(data), verify=self.verify) as response:
                response.raise_for_status()
                content = json.loads(response.content)
                self.auth_token = content['auth_token']

        return {'Authorization': 'Bearer {auth_token}'.format(
            auth_token=self.auth_token)}

    def _accesspoint_url(self, endpoint: str = '', node: int = None):
        """ convert an endpoint into a full URL for an API Call

        Pass in a sub-url endpoint and this will convert it into a full URL.
        You can request a specific host index if desired.

        Parameters:
        -----------

        endpoint (str) : API endpoint you are trying to access

        node (int) : If not None, then the API call should be directed to a specific
            node index from the list of hosts, as opposed to the generic accesspoint.
            This allows things like ping confirmation on specific host.

        """
        if node is None:
            target = self.accesspoint
        else:
            target = self._node_address(node)

        return "https://{accesspoint}/{endpoint}".format(
            accesspoint=self.accesspoint, endpoint=endpoint)

    def _node_address(self, node: int = 0):
        """ get the ip address from the node for the node index """
        node = self.hosts[node]
        if 'address' in node:
            return node['address']
        elif 'ssh' in node:
            return node['ssh']['address']
        elif 'winrm' in node:
            return node['winrm']['address']



"""

Ansible provisioner

"""

from typing import Any

from configerus.loaded import LOADED_KEY_ROOT

from mirantis.testing.metta.plugin import Factory, Type
from mirantis.testing.metta.environment import Environment

from .provisioner import AnsibleProvisionerPlugin, ANSIBLE_PROVISIONER_CONFIG_LABEL
from .cli import AnsibleCliPlugin

METTA_ANSIBLE_PROVISIONER_PLUGIN_ID = 'metta_ansible'
""" Ansible provisioner plugin id """


@Factory(type=Type.PROVISIONER, plugin_id=METTA_ANSIBLE_PROVISIONER_PLUGIN_ID)
def metta_plugin_factory_provisioner_ansible(
        environment: Environment, instance_id: str = "", label: str = ANSIBLE_PROVISIONER_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
    """ create an metta provisionersss dict plugin """
    return AnsibleProvisionerPlugin(environment, instance_id, label, base)


METTA_ANSIBLE_CLI_PLUGIN_ID = 'metta_ansible'
""" cli plugin_id for the info plugin """


@Factory(type=Type.CLI, plugin_id=METTA_ANSIBLE_CLI_PLUGIN_ID)
def metta_ansible_factory_cli_ansible(
        environment: Environment, instance_id: str = ''):
    """ create an info cli plugin """
    return AnsibleCliPlugin(environment, instance_id)


""" SetupTools EntryPoint METTA BootStrapping """


def bootstrap(environment: Environment):
    """ METTA_Ansible bootstrap

    We dont't take any action.  Our purpose is to run the above factory
    decorator to register our plugin.

    """
    pass



import os.path
import logging
import pkg_resources

from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT
from mirantis.testing.metta.environment import Environment

logger = logging.getLogger("metta_mirantis.presets")

METTA_PRESET_CONFIG_LABEL = "mirantis"
""" This config label will be loaded to interpret config in add_preset_config """
METTA_CONFIG_SOURCE_INSTANCE_ID_PREFIX = "metta_mirantis-preset"
""" All config that we add here will have its source instance_id prefixed with this """
METTA_CONFIG_CONFIG_PRESET_BASE = 'presets'
""" preset configuration will be found under this config .get() key """
METTA_PRESET_DEFAULT_PRIORITY = 60
""" Default priority to use for preset config sources """
METTA_PRESET_PACKAGE = 'mirantis.testing.metta_mirantis'


def preset_config():
    """ List of available presets that you can include in load("metta").get('presets')
        (key, preset_root_path):
            key: config key that will find the preset value under the base value
                from METTA_CONFIG_CONFIG_PRESET_BASE
            path: preset root path that should contains a path matching the value
                which will be added as a source config path
            priority delta: preset relative priority compared to other presets
    """
    return [
        ("variation", pkg_resources.resource_filename(
            METTA_PRESET_PACKAGE, "config/variation"), 0),
        ("cluster", pkg_resources.resource_filename(
            METTA_PRESET_PACKAGE, "config/cluster"), 1),
        ("platform", pkg_resources.resource_filename(
            METTA_PRESET_PACKAGE, "config/platform"), 1),
        ("release", pkg_resources.resource_filename(
            METTA_PRESET_PACKAGE, "config/release"), 1)
    ]


def add_preset_config(environment: Environment,
                      priority=METTA_PRESET_DEFAULT_PRIORITY):
    """ Read metta config and interpret it for modifying the config

    Interprests config.load("metta") to determine if any `presets` config
    sources should be added.

    The following presets are currently available:

    1. variation : see ./config/variations : kind of access to a larger config
        pattern that we use at Mirantis.  An example is `lat` which refers to
        using launchpad & terraform with our common config.

    2. cluster : @see ./config/cluster : config for different cluster sizes
    3. release : @see ./config/release : config for different mirantis product
        releases
    4. platform : @see ./config/platform : config for different os platforms

    In all cases, the pattern is that a key from config.load("metta") is
    used to pull a preset name, and the matching preset subpath is added to the
    config object as a path source.

    The presets for #2, #3 and # are focused on overriding `variables` (label)
    which should be consumed by whichever variation you use.

    Parameter:
    ----------

    config (configerus.config.Config) : a configerus Config object which will
        have config sources added

    priority (int) : optionally override the source plugin priority.

    Throws:
    -------

    Can throw a KeyError if a preset id requested that does not exist

    """

    # Now we start to process `presets`
    # To do this we load the `metta` config label and look for preset
    # keys in the loaded config.
    # For each preset key found we try to add a path config source for the
    # value
    try:
        metta_config = environment.config.load(METTA_PRESET_CONFIG_LABEL)
    except KeyError:
        logger.debug("metta Presets found no usable config")
        return

    for preset in preset_config():
        # For each of the available presets, look in the loaded metta
        # config for values matching the preset key.  if the key is found then
        # we try to match a path in the metta/config/{key}/{value}
        # and add the path as a config source.

        (preset_key, preset_root_path, preset_priority_delta) = preset
        preset_priority = priority + preset_priority_delta
        preset_value = metta_config.get(
            [METTA_CONFIG_CONFIG_PRESET_BASE, preset_key], default='')

        if preset_value:
            preset_instance_id = "{}-{}-{}".format(
                METTA_CONFIG_SOURCE_INSTANCE_ID_PREFIX, preset_key, preset_value.replace('/', '_'))
            # quick check to see if we've already added this preset.
            if not environment.config.has_source(preset_instance_id):
                # build a preset config path and add it as a source if it
                # exists
                preset_full_path = os.path.join(preset_root_path, preset_value)
                if os.path.isdir(preset_full_path):
                    logger.debug(
                        "Using metta preset {}:{} => {}".format(
                            preset_key, preset_value, preset_full_path))
                    environment.config.add_source(
                        PLUGIN_ID_SOURCE_PATH,
                        preset_instance_id,
                        preset_priority).set_path(preset_full_path)
                else:
                    raise KeyError(
                        "metta doesn't have a preset '%s:%s'",
                        preset_key,
                        preset_value)

        else:
            logger.debug("No metta preset selected for {}".format(preset_key))



from typing import Any, List
import logging
import os
import subprocess
import json
from enum import Enum
from datetime import datetime

import kubernetes

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.workload import WorkloadBase

logger = logging.getLogger('metta.contrib.kubernetes.workload.helm')


class HelmWorkloadPlugin(WorkloadBase):
    """ Helm client for interacting with a kube client """


KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL = 'kubernetes'
""" default config label used to load the workload """
KUBERNETES_HELM_WORKLOAD_CONFIG_BASE = 'workload.helm'
""" default config key for creating a workload object """

KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VERSION = 'version'
""" config key for helm version """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_REPOS = 'repos'
""" config key for helm repos dict which need to be added """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_CHART = 'chart'
""" config key for helm chart, either local path or http(s) """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUES = 'values'
""" config key for helm chart values """
KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_NAMESPACE = 'namespace'
""" config key for namespace to install to """

KUBERNETES_HELM_WORKLOAD_DEFAULT_NAMESPACE = 'default'
""" default namespace to install to if no namespace was passed """


class KubernetesHelmWorkloadPlugin(WorkloadBase):
    """ Kubernetes workload class """

    def __init__(self, environment, instance_id,
                 label: str = KUBERNETES_HELM_WORKLOAD_CONFIG_LABEL, base: Any = KUBERNETES_HELM_WORKLOAD_CONFIG_BASE):
        """ Run the super constructor but also set class properties

        This implements the args part of the client interface.

        Here we expect to receiv config pointers so that we can determine what
        helm workload to apply, and what values will be needed.

        Parameters:
        -----------

        config_file (str): String path to the kubernetes config file to use

        """
        WorkloadBase.__init__(self, environment, instance_id)

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

    def create_instance(self, fixtures: Fixtures):
        """ Create a workload instance from a set of fixtures

        Parameters:
        -----------

        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a kubernetes client plugin.

        """

        try:
            client = fixtures.get_plugin(
                type=Type.CLIENT, plugin_id='metta_kubernetes')
        except KeyError as e:
            raise NotImplementedError(
                "Workload could not find the needed client: {}".format('metta_kubernetes'))

        workload_config = self.environment.config.load(self.config_label)

        kubeconfig = client.config_file

        namespace = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_NAMESPACE], default=KUBERNETES_HELM_WORKLOAD_DEFAULT_NAMESPACE)

        chart = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_CHART])

        values = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUES], default={})

        repos = workload_config.get(
            [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_REPOS], default={})

        return KubernetesHelmV3WorkloadInstance(
            kubeconfig, namespace, self.instance_id, repos, chart, values)

    def info(self):
        """ Return dict data about this plugin for introspection """
        workload_config = self.environment.config.load(self.config_label)

        return {
            'workload': {
                'deployment': {
                    'repos': workload_config.get(
                        [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_REPOS]),
                    'chart': workload_config.get(
                        [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_CHART]),
                    'values': workload_config.get(
                        [self.config_base, KUBERNETES_HELM_WORKLOAD_CONFIG_KEY_VALUES])
                },
                'required_fixtures': {
                    'kubernetes': {
                        'type': Type.CLIENT.value,
                        'plugin_id': 'metta_kubernetes'
                    }
                }
            }
        }


class KubernetesHelmV3WorkloadInstance:

    def __init__(self, kubeconfig, namespace, name, repos, chart, values):
        """

        Parameters:
        -----------

        kubeconfig (str) : Path to the kubeinfo file

        repos (dict[str, str]) : name->path for repos to add

        chart (str) : local or http(s) path to a chart

        values (dict[str, Any]) : helm chart values

        """

        self.kubeconfig = kubeconfig
        self.name = name
        self.namespace = namespace
        self.repos = repos
        self.chart = chart
        self.values = values

        self.bin = 'helm'
        self.working_dir = '.'

    def apply(self, wait: bool = True, debug: bool = False):
        """ Apply the helm chart """

        for repo_name, repo_url in self.repos.items():
            self._run(cmd=['repo', 'add', repo_name, repo_url])

        if len(self.values):
            # turn the values dict into '--set a=A,b=B,c=C'
            values = ['--set', ','.join(['{}={}'.format(name, value)
                                         for (name, value) in self.values.items()])]
        else:
            values = []

        cmd = ['install']
        cmd += [self.name, self.chart]

        if wait:
            cmd += ['--wait']
        if debug:
            cmd += ['--debug']

        cmd += values

        try:
            self._run(cmd=cmd)
        except Exception as e:
            raise RuntimeError(
                'Helm failed to install the relese: {}'.format(e)) from e

    def list(self, all: bool = False, failed: bool = False,
             deployed: bool = False, pending: bool = False):
        """ List all releases - Not instances specific but still useful """
        cmd = ['list', '--output={}'.format('json')]

        if all:
            cmd += ['--all']
        elif deployed:
            cmd += ['--deployed']
        elif failed:
            cmd += ['--failed']
        elif pending:
            cmd += ['--pending']

        list_str = self._run(cmd=cmd, return_output=True)
        if list_str:
            return json.loads(list_str)

        return []

    def destroy(self):
        """ remove an installed helm release """
        self._run(cmd=['uninstall', self.name])

    def test(self):
        """ test an installed helm release """
        self._run(cmd=['test', self.name])

    def status(self):
        """ status of an installed helm release """
        return HelmReleaseStatus(json.loads(
            self._run(cmd=['status', self.name, '--output', 'json'], return_output=True)))

    def _run(self, cmd: List[str], return_output: bool = False):
        """ run a helm v3 command """

        cmd = [self.bin,
               '--kubeconfig={}'.format(self.kubeconfig),
               '--namespace={}'.format(self.namespace)] + cmd

        if return_output:
            logger.debug(
                "running launchpad command with output capture: %s",
                " ".join(cmd))
            exec = subprocess.run(
                cmd,
                cwd=self.working_dir,
                shell=False,
                stdout=subprocess.PIPE)
            exec.check_returncode()
            return exec.stdout.decode('utf-8')
        else:
            logger.debug("running launchpad command: %s", " ".join(cmd))
            exec = subprocess.run(
                cmd,
                cwd=self.working_dir,
                check=True,
                text=True)
            exec.check_returncode()


class Status(Enum):
    """ A Helm Status enum """
    DEPLOYED = 'deployed'
    """ Helm Release has been deployed """


class HelmReleaseStatus:
    """ interpreted helm release status """

    def __init__(self, status_reponse: dict):
        """ interpret status from a status response """
        self.name = status_reponse['name']
        self.version = status_reponse['version']
        self.namespace = status_reponse['namespace']

        # self.first_deployed = datetime.strptime(status_reponse['info']['first_deployed'], '%Y-%m-%dT%H:%M:%S.%fZ')
        # self.last_deployed = datetime.strptime(status_reponse['info']['last_deployed'], '%Y-%m-%dT%H:%M:%S.%fZ')
        self.deleted = status_reponse['info']['deleted']
        self.description = status_reponse['info']['description']
        self.status = Status(status_reponse['info']['status'])
        self.notes = status_reponse['info']['notes']

        self.config = status_reponse['config']

        self.manifect = status_reponse['manifest']



"""

metta Mirantis


"""

import logging
from typing import List, Dict

from mirantis.testing.metta.plugin import Factory, Type
from mirantis.testing.metta.environment import Environment

from .common import add_common_config
from .presets import add_preset_config

from .mke_client import MKEAPIClientPlugin, MKEAPICliPlugin, METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID
from .msr_client import MSRAPIClientPlugin, MSRAPICliPlugin, METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID

""" Plugin factories """


@Factory(type=Type.CLIENT, plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID)
def metta_mirantis_plugin_factory_client_mke(
        environment: Environment, instance_id: str = "", accesspoint: str = '', username: str = '', password: str = '', hosts: List[Dict] = []):
    """ Create a Mirantis MKE API Client """
    return MKEAPIClientPlugin(environment, instance_id, accesspoint=accesspoint,
                              username=username, password=password, hosts=hosts)


@Factory(type=Type.CLI, plugin_id=METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID)
def metta_terraform_factory_cli_mke(
        environment: Environment, instance_id: str = ''):
    """ create an MKE cli plugin """
    return MKEAPICliPlugin(environment, instance_id)


@Factory(type=Type.CLIENT, plugin_id=METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID)
def metta_mirantis_plugin_factory_client_msr(
        environment: Environment, instance_id: str = '', accesspoint: str = '', username: str = '', password: str = '', hosts: List[Dict] = []):
    """ Create a Mirantis MSR API Client """
    return MSRAPIClientPlugin(environment, instance_id, accesspoint=accesspoint,
                              username=username, password=password, hosts=hosts)


@Factory(type=Type.CLI, plugin_id=METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID)
def metta_terraform_factory_cli_msr(
        environment: Environment, instance_id: str = ''):
    """ create an MSR cli plugin """
    return MSRAPICliPlugin(environment, instance_id)


""" METTA bootstraps that we will use on config objects """


def bootstrap_common(environment: Environment):
    """ metta configerus bootstrap

    Add some Mirantis specific config options for presets

    Additional config sources may be added based on config.load("metta")
    which will try to add config for metta `presets`
    This gives access to "variation", "cluster", "platform" and "release"
    presets.  You can dig deeper, but it might implicitly make sense if you look
    at the config folder of this module.

    @see .common.add_common_config() for details

    """
    add_common_config(environment)


def bootstrap_presets(environment: Environment):
    """ metta configerus bootstrap

    Add some Mirantis specific config options for presets

    Additional config sources may be added based on config.load("metta")
    which will try to add config for metta `presets`
    This gives access to "variation", "cluster", "platform" and "release"
    presets.  You can dig deeper, but it might implicitly make sense if you look
    at the config folder of this module.

    @see .presets.add_preset_config() for details

    """
    add_preset_config(environment)


__all__ = [
    bootstrap_common,
    bootstrap_presets,
    METTA_MIRANTIS_CLIENT_MKE_PLUGIN_ID,
    METTA_MIRANTIS_CLIENT_MSR_PLUGIN_ID]



"""

Docker workloads plugin

"""
from typing import Dict, Any
import logging

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.fixtures import Fixtures
from mirantis.testing.metta.workload import WorkloadBase, WorkloadInstanceBase

logger = logging.getLogger('metta.contrib.docker.workload.run')

DOCKER_RUN_WORKLOAD_CONFIG_LABEL = 'docker'
""" Configerus label for retrieving docker run workloads """
DOCKER_RUN_WORKLOAD_CONFIG_BASE = 'workload.run'
""" Configerus get base for retrieving the default run workload """


class DockerPyRunWorkloadPlugin(WorkloadBase):
    """ Docker Run workload class for the DockerPy """

    def __init__(self, environment: Environment, instance_id: str,
                 label: str = DOCKER_RUN_WORKLOAD_CONFIG_LABEL, base: Any = DOCKER_RUN_WORKLOAD_CONFIG_BASE):
        """ Run the super constructor but also set class properties

        Parameters:
        -----------

        outputs (Dict[Dict]) : pass in a dictionary which defines outputs that
            should be returned

        clients (Dict[Dict]) : pass in a dictionary which defines which clients
            should be requested when working on a provisioner

        """
        WorkloadBase.__init__(self, environment, instance_id)

        logger.info("Preparing Docker run setting")

        self.config_label = label
        """ configerus load label that should contain all of the config """
        self.config_base = base
        """ configerus get key that should contain all tf config """

    def create_instance(self, fixtures: Fixtures):
        """ Create a workload instance from a set of fixtures

        Parameters:
        -----------

        fixtures (Fixtures) : a set of fixtures that this workload will use to
            retrieve a docker client plugin.

        """

        client = fixtures.get_plugin(
            type=Type.CLIENT, plugin_id='metta_docker')

        run_config = self.environment.config.load(self.config_label)
        """ get a configerus LoadedConfig for the docker run label """

        run = run_config.get([self.config_base, 'run'])

        return DockerRunWorkloadInstance(client, run)

    def info(self):
        """ Return dict data about this plugin for introspection """

        run_config = self.environment.config.load(self.config_label)
        """ get a configerus LoadedConfig for the docker run label """

        return {
            'workload': {
                'run': {
                    'run': run_config.get([self.config_base, 'run'])
                },
                'required_fixtures': {
                    'docker': {
                        'type': Type.CLIENT.value,
                        'plugin_id': 'metta_docker'
                    }
                }
            }
        }


class DockerRunWorkloadInstance(WorkloadInstanceBase):
    """ A workload isntance for a docker run """

    def __init__(self, client, run):
        self.client = client
        self.run = run

    def apply(self):
        """ Run the workload

        @NOTE Needs a docker client fixture to run.  Use .set_fixtures() first

        """

        assert 'image' in self.run, "Run command had no image"

        self.run = self.client.containers.run(**self.run)
        return self.run



import logging
from typing import Dict, Any
import json

from configerus.loaded import LOADED_KEY_ROOT
from configerus.plugin import Type as ConfigerusType
from configerus.contrib.files import PLUGIN_ID_SOURCE_PATH
from configerus.contrib.dict import PLUGIN_ID_SOURCE_DICT

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase

logger = logging.getLogger('metta.cli.config')


class ConfigCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands """
        return {
            'config': ConfigGroup(self.environment)
        }


class ConfigGroup():

    def __init__(self, environment: Environment):
        self.environment = environment

    def plugins(self, plugin_id: str = '', instance_id: str = '',
                type: str = ''):
        """ List configerus plugins """
        list = [{
            'type': instance.type.value,
            'plugin_id': instance.plugin_id,
            'instance_id': instance.instance_id,
            'priority': instance.priority,
        } for instance in self.environment.config.plugins.get_instances(plugin_id=plugin_id, instance_id=instance_id, type=type)]

        return json.dumps(list, indent=2)

    def sources(self, plugin_id: str = '',
                instance_id: str = '', deep: bool = False):
        """ List configerus sourceS """
        list = []
        for instance in self.environment.config.plugins.get_instances(
                plugin_id=plugin_id, instance_id=instance_id, type=ConfigerusType.SOURCE):
            source = {
                'type': instance.type.value,
                'plugin_id': instance.plugin_id,
                'instance_id': instance.instance_id,
                'priority': instance.priority,
            }

            if deep:
                if instance.plugin_id == PLUGIN_ID_SOURCE_PATH:
                    source['path'] = instance.plugin.path
                if instance.plugin_id == PLUGIN_ID_SOURCE_DICT:
                    source['data'] = instance.plugin.data

            list.append(source)

        return json.dumps(list, indent=2, default=serialize_last_resort)

    def loaded(self, raw: bool = False):
        """ List loaded config labels """
        loaded = self.environment.config.loaded
        value = list(loaded)
        if raw:
            return value
        else:
            return json.dumps(value)

    def get(self, label: str, key: str = LOADED_KEY_ROOT):
        """ Retrieve configuration from the config object

        USAGE:

            metta config get {label} [{key}]


        """
        try:
            loaded = self.environment.config.load(label)
        except KeyError as e:
            return "Could not find the config label '{}'".format(label)

        try:
            value = loaded.get(key)
        except KeyError as e:
            return "Could not find the config key '{}'".format(key)

        return json.dumps(value, indent=2, default=serialize_last_resort)

    def format(self, data: str,
               default_label: str = 'you did not specify a default', raw: bool = False):
        """ Format a target string using config templating """

        try:
            value = self.environment.config.format(
                data=data, default_label=default_label)
        except Exception as e:
            return "Error occured in formatting: '{}'".format(e)

        if raw:
            return value
        else:
            return json.dumps(value, indent=2, default=serialize_last_resort)


def serialize_last_resort(X):
    """ last attempt at serializing """
    return "{}".format(X)



"""

A provisioner that combines multiple provisioners

"""
import logging
from typing import Dict, List, Any

from configerus.loaded import LOADED_KEY_ROOT
from mirantis.testing.metta.plugin import METTAPlugin, Type
from mirantis.testing.metta.fixtures import Fixtures, Fixture, UCCTFixturesPlugin
from mirantis.testing.metta.provisioner import ProvisionerBase

logger = logging.getLogger('metta.contrib.provisioner:combo')

COMBO_PROVISIONER_CONFIG_LABEL = 'provisioner'
""" Configerus label for loading config to set up this provisioner plugin """
COMBO_PROVISIONER_CONFIG_BACKENDS_KEY = 'backends'
""" Config key for backends list """


class ComboProvisionerPlugin(ProvisionerBase, UCCTFixturesPlugin):
    "Combo Provisioner plugin class"

    def __init__(self, environment, instance_id,
                 label: str = COMBO_PROVISIONER_CONFIG_LABEL, base: Any = LOADED_KEY_ROOT):
        """ Run the super constructor but also set class properties

        Interpret provided config and configure the object with all of the needed
        pieces for executing terraform commands

        """
        ProvisionerBase.__init__(self, environment, instance_id)
        UCCTFixturesPlugin.__init__(self)

        logger.info("Preparing Combo provisioner")

        try:
            self.combo_config = self.environment.config.load(label)
        except KeyError as e:
            raise ValueError(
                "Combo provisioner could not find any configurations: {}".format(e)) from e

        self.backends = self.combo_config.get(
            [base, COMBO_PROVISIONER_CONFIG_BACKENDS_KEY])
        if not isinstance(self.backends, list):
            raise ValueError(
                "Combo provisioner could not understand the backend list.")

        for backend in self.backends:
            backend_instance_id = backend['instance_id']
            fixture = self.environment.fixtures.get_fixture(
                type=Type.PROVISIONER, instance_id=backend_instance_id)

            if hasattr(backend, "priority"):
                fixture.priority = backend['priority']

            self.fixtures.add_fixture(fixture)

    def _get_ordered_backend_fixtures(self, high_to_low: bool = False):
        """ helper to get the sorted backend fixtures from lowest priority to highest, reversed if requested """
        backend_fixtures = self.fixtures.get_fixtures(
            type=Type.PROVISIONER).to_list()
        if not high_to_low:
            backend_fixtures.reverse
        return backend_fixtures

    def info(self):
        """ return structured data about self. """

        backends_info = []
        for fixture in self._get_ordered_backend_fixtures():
            backend_info = {
                'fixture': {
                    'type': fixture.type.value,
                    'instance_id': fixture.instance_id,
                    'plugin_id': fixture.plugin_id
                }
            }

            plugin = fixture.plugin
            if hasattr(plugin, 'info'):
                backend_info.update(plugin.info())

            if not 'fixture' in backend_info:
                backend_info

            backends_info.append(backend_info)

        return {
            'backends': backends_info
        }

    def prepare(self, label: str = '', base: str = ''):
        """ Prepare the provisioner to apply resources """
        for backend_fixture in self._get_ordered_backend_fixtures():
            logger.info(
                "--> running backend prepare: [Low->High] {}".format(backend_fixture.instance_id))
            backend_fixture.plugin.prepare()

    def apply(self):
        """ bring a cluster to the configured state """
        for backend_fixture in self._get_ordered_backend_fixtures():
            logger.info(
                "--> running backend apply: [Low->High] {}".format(backend_fixture.instance_id))
            backend_fixture.plugin.apply()

    def destroy(self):
        """ remove all resources created for the cluster """
        for backend_fixture in self._get_ordered_backend_fixtures(
                high_to_low=True):
            logger.info(
                "--> running backend destroy: [High->Low] {}".format(backend_fixture.instance_id))
            backend_fixture.plugin.destroy()

    """ Fixture management """

    def get_fixtures(self, type: Type = None, instance_id: str = '',
                     plugin_id: str = '') -> Fixtures:
        """ retrieve any matching fixtures from any of the backends """
        matches = Fixtures()
        for backend_fixture in self._get_ordered_backend_fixtures():
            plugin = backend_fixture.plugin
            if hasattr(plugin, 'get_fixtures'):
                for match in plugin.get_fixtures(
                        type=type, plugin_id=plugin_id, instance_id=instance_id).to_list():
                    matches.add_fixture(match)
        return matches

    def get_fixture(self, type: Type = None, instance_id: str = '',
                    plugin_id: str = '', exception_if_missing: bool = True) -> Fixture:
        """ retrieve the first matching fixture fomr backend in high-to-low order """

        for backend_fixture in self._get_ordered_backend_fixtures(
                high_to_low=True):
            plugin = backend_fixture.plugin
            if hasattr(plugin, 'get_fixture'):
                fixture = plugin.get_fixture(
                    type=type,
                    plugin_id=plugin_id,
                    instance_id=instance_id,
                    exception_if_missing=exception_if_missing)
                if fixture is not None:
                    return fixture

        if exception_if_missing:
            raise KeyError("No matching fixture was found")

    def get_plugin(self, type: Type = None, plugin_id: str = '',
                   instance_id: str = '', exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixtures """
        logger.info(
            "{}:execute: get_plugin({})".format(
                self.instance_id,
                type.value))
        fixture = self.get_fixture(
            type=type, plugin_id=plugin_id, instance_id=instance_id, exception_if_missing=exception_if_missing)

        if fixture is not None:
            return fixture.plugin

    def get_provisioner(self, plugin_id: str = '', instance_id: str = '',
                        exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture provisioner """
        return self.get_plugin(type=Type.PROVISIONER, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)

    def get_output(self, plugin_id: str = '', instance_id: str = '',
                   exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture outputs """
        return self.get_plugin(type=Type.OUTPUT, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)

    def get_client(self, plugin_id: str = '', instance_id: str = '',
                   exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture clients """
        return self.get_plugin(type=Type.CLIENT, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)

    def get_workload(self, plugin_id: str = '', instance_id: str = '',
                     exception_if_missing: bool = True) -> METTAPlugin:
        """ Retrieve one of the passed in fixture workloads """
        return self.get_plugin(type=Type.WORKLOAD, plugin_id=plugin_id,
                               instance_id=instance_id, exception_if_missing=exception_if_missing)



"""

METTA Plugin definition and management

@NOTE we would like to type hint the 'environment' variable but it would
    create a circular import. This is not totally avoidable as plugin
    holds an Environment, but an Environments creates the plugin

"""
import logging
import functools
from enum import Enum, unique

logger = logging.getLogger('metta.plugin')

METTA_PLUGIN_CONFIG_KEY_PLUGIN = 'plugin'
""" configerus .get() key for a single plugin """
METTA_PLUGIN_CONFIG_KEY_PLUGINID = 'plugin_id'
""" configerus .get() key for plugin_id """
METTA_PLUGIN_CONFIG_KEY_INSTANCEID = 'instance_id'
""" configerus .get() key for plugin_id """
METTA_PLUGIN_CONFIG_KEY_TYPE = 'type'
""" configerus .get() key for plugin type """
METTA_PLUGIN_CONFIG_KEY_ARGUMENTS = 'arguments'
""" configerus .get() key for plugin arguments """
METTA_PLUGIN_CONFIG_KEY_PRIORITY = 'priority'
""" will use this Dict key assign an instance a priority when it is created. """
METTA_PLUGIN_CONFIG_KEY_CONFIG = 'config'
""" will use this Dict key as additional config """
METTA_PLUGIN_CONFIG_KEY_VALIDATORS = 'validators'
""" will use this Dict key from the output config to decide what validators to apply to the plugin """


METTA_PLUGIN_VALIDATION_JSONSCHEMA = {
    METTA_PLUGIN_CONFIG_KEY_INSTANCEID: {'type': 'string'}
}
""" json schema validation definition for a plugin """


class METTAPlugin():
    """ Base metta Plugin which all plugins can extend """

    def __init__(self, environment: object, instance_id: str):
        """
        Parameters:
        -----------

        environment (Environment) : Environment object in which this plugin lives.

        instance_id (str) : instance id for the plugin.  A unique identifier which
            the plugin can use for naming and introspective identification.

        """
        self.environment = environment
        self.instance_id = instance_id


@unique
class Type(Enum):
    """ Enumerator to match plugin types to plugin labels """
    CLIENT = 'metta.plugin.client'
    """ Plugins which interact with clusters """
    SOURCE = 'metta.plugin.configsource'
    """ A Config source handler """
    OUTPUT = 'metta.plugin.output'
    """ A Config output handler """
    PROVISIONER = 'metta.plugin.provisioner'
    """ A cluster provisioner plugin """
    WORKLOAD = 'metta.plugin.workload'
    """ Plugins which use clients/provisioners to apply a workload to a cluster """
    UTILITY = 'metta.plugins.utility'
    """ plugins that are used for tooling around testing, but don't directly participate in testing """
    CLI = 'metta.plugin.cli'
    """ Plugins extend the metta cli """

    def from_string(type_string: str) -> 'Type':
        """ Try to offer some flexibility when defining a type

        METTA Plugin enum keys are upper case keys such as "OUTPUT" or "CLIENT"
        and values are in a long form 'metta.plugin.provisioner'.  Both can be
        difficult to use, especially for config ources which can't
        programmatically access the enum.

        This function allows some flexibility in naming.

        Parameters:
        -----------

        type_string (str) : a string attempt to identify a plugin.  It can be
            an any-case form of the Type key, the full string value of a Type
            instance, or the part after 'metta.plugin.' of the value.

        Returns:
        --------

        a Type instance matching the passed string

        Raises:
        -------

        KeyError if the passed argument could not be matched.

        """
        if type is None:
            raise ValueError(
                "Cannot determine type of plugin sa you passed a None value")

        try:
            return Type(type_string)
        except ValueError:
            pass
        try:
            return Type[type_string.upper()]
        except (KeyError, AttributeError):
            pass
        try:
            return Type('metta.plugin.{}'.format(type_string))
        except ValueError:
            pass

        raise KeyError(
            "Could not identify METTA plugin type requested '{}'".format(type_string))


class Factory():
    """ Python decorator class for metta Plugin factories

    This class should be used to decorate any function which is meant to be a
    factory for metta plugins.

    If you are writing a plugin factory, decorate it with this class, provide
    the factory type and id values, and then the factory will be avaialble to
    other code.

    If you are trying to get an instance of a plugin, then create an instance of
    this class and use the create() method

    """

    registry = {}
    """ A list of all of the registered factory functions """

    def __init__(self, type: Type, plugin_id: str):
        """ register the decoration

        Parameters:
        -----------

        type (Type) : Plugin type which the factory created

        plugin_id (str) : unique identifier for the plugin which this factory
            will create.  This will be used on registration and the matching
            value is used on construction.

        """
        logger.debug(
            "Plugin factory registered `%s:%s`",
            type.value,
            plugin_id)
        self.plugin_id = plugin_id
        self.type = type

        if not self.type.value in self.registry:
            self.registry[self.type.value] = {}

    def __call__(self, func):
        """ Decorator factory wrapping function

        Returns:
        --------

        Decorated construction function(config: Config)

        """
        functools.update_wrapper(self, func)

        def wrapper(environment: object, instance_id: str, *args, **kwargs):
            logger.debug(
                "plugin factory exec: %s:%s",
                self.type.value,
                self.plugin_id)
            plugin = func(
                environment=environment,
                instance_id=instance_id,
                *args,
                **kwargs)
            if not isinstance(plugin, METTAPlugin):
                logger.warn(
                    "plugin factory did not return an instance of metta Plugin `{}:{}`".format(
                        self.type.value, self.plugin_id))

            return plugin

        self.registry[self.type.value][self.plugin_id] = wrapper
        return wrapper

    def create(self, environment: object, instance_id: str, *args, **kwargs):
        """ Get an instance of a plugin as created by the decorated

        Parameters:

        environment (Environment) : Environment object in which this plugin lives.

        instance_id (str) : instance id for the plugin.  A unique identifier which
            the plugin can use for naming and introspective identification.

        """
        try:
            factory = self.registry[self.type.value][self.plugin_id]
        except KeyError:
            raise NotImplementedError(
                "METTA Plugin instance '{}:{}' has not been registered.".format(
                    self.type.value, self.plugin_id))
        except Exception as e:
            raise Exception(
                "Could not create Plugin instance '{}:{}' as the plugin factory produced an exception".format(
                    self.type.value, self.plugin_id)) from e

        return factory(environment=environment,
                       instance_id=instance_id, *args, **kwargs)



from typing import Any

from mirantis.testing.metta.plugin import Factory, Type
from mirantis.testing.metta.environment import Environment

from .sonobuoy import SonobuoyWorkloadPlugin, METTA_PLUGIN_ID_SONOBUOY_WORKLOAD, SONOBUOY_WORKLOAD_CONFIG_LABEL, SONOBUOY_WORKLOAD_CONFIG_BASE
from .cli import SonobuoyCliPlugin, METTA_PLUGIN_ID_SONOBUOY_cli


@Factory(type=Type.WORKLOAD,
         plugin_id=METTA_PLUGIN_ID_SONOBUOY_WORKLOAD)
def metta_plugin_factory_workload_sonobuoy(
        environment: Environment, instance_id: str = '', label: str = SONOBUOY_WORKLOAD_CONFIG_LABEL, base: Any = SONOBUOY_WORKLOAD_CONFIG_BASE):
    """ create an metta kubernetes spec workload plugin """
    return SonobuoyWorkloadPlugin(
        environment, instance_id, label=label, base=base)


@Factory(type=Type.CLI, plugin_id=METTA_PLUGIN_ID_SONOBUOY_cli)
def metta_plugin_factory_cli_sonobuoy(
        environment: Environment, instance_id: str = ''):
    """ create an sonobuoy cli plugin """
    return SonobuoyCliPlugin(environment, instance_id)


""" SetupTools EntryPoint METTA BootStrapping """


def bootstrap(environment: Environment):
    """ METTA_Sonobuoy bootstrap

    We dont't take any action.  Our purpose is to run the above factory
    decorators to register our plugins.

    """
    pass


# Export imported contants to make it easier for consumers to import
__all__ = [
    bootstrap,
    metta_plugin_factory_workload_sonobuoy,
    metta_plugin_factory_cli_sonobuoy,
    METTA_PLUGIN_ID_SONOBUOY_WORKLOAD,
    SONOBUOY_WORKLOAD_CONFIG_LABEL,
    SONOBUOY_WORKLOAD_CONFIG_BASE]




import logging
import itertools
from typing import List
from threading import Thread

from mirantis.testing.metta.environment import Environment

from .preset import combination_config

logger = logging.getLogger("metta_mirantis.matrix")


class MatrixRun:
    """ Matrix decorator that will run a method for all matrix combinations

    Expects to decorate a method: func(config:Config, provisioner)

    """

    def __init__(self, platforms: List[str] = [],
                 clusters: List[str] = [], releases: List[str] = []):
        """ register the decoration """
        logger.debug("MatrixTestRun")

        self.matrix = list(itertools.product(*[
            platforms,
            clusters,
            releases
        ]))

    def __call__(self, func):
        """ call the decorated function

        Returns:

        wrapped function(config: Config)
        """

        def matrix_wrap(environment: Environment):
            logger.debug("Matrix test run exec")

            variables_config = config.load("variables")

            i = 0
            for combination in self.matrix:

                metta_dict = {}
                if combination[0]:
                    metta_dict["platform"] = combination[0]
                if combination[1]:
                    metta_dict["cluster"] = combination[1]
                if combination[2]:
                    metta_dict["release"] = combination[2]

                id = str("-".join(metta_dict.values())).replace("/", "-")
                logger.debug("MatrixTestRun::Combination: %s", id)

                metta_dict["id"] = "{}-{}".format(
                    variables_config.get("id"), id).replace(
                    "/", "-")

                combination_config = config.copy()
                combination_config.add_source(metta.SOURCE_DICT, "combination", config.default_priority() + 1).set_data({
                    "metta": metta_dict,
                    "variables": {
                        "id": "{}-{}".format(func.__code__.co_name, id).replace("/", "-"),
                        "resource_prefix": "{}-matrix{}".format(func.__code__.co_name, i).replace("/", "-").replace("_", "-")
                    }
                })

                # tell metta to add variation/platform/release/cluster
                add_preset_config(combination_config)
                # build a new metta provisioner now that we have added config
                combination_provisioner = metta.new_provisioner_from_config(
                    combination_config, "matrix_provisioner")

                thread = ProvisionerUpFuncCallThread(
                    id, combination_config, combination_provisioner, func)
                thread.start()

                i += 1

        return matrix_wrap


class ProvisionerUpFuncCallThread(Thread):

    # def __init__(self,id, config, provisioner, func):
    def __init__(self, id, config, provisioner, func):
        self.logger = logger.getChild(
            "{}:matrix:{}".format(
                func.__code__.co_name, id))
        self.id = id
        self.config = config
        self.provisioner = provisioner
        self.func = func

        Thread.__init__(self)

    def run(self):
        self.logger.debug("RUNNING " + self.id)

        conf = self.config.load("config")
        provisioner_config = self.config.load("provisioner")

        try:
            self.logger.info(
                "Preparing the testing cluster using the provisioner")
            self.provisioner.prepare()
        except Exception as e:
            raise Exception("Provisioner failed to prepare()") from e
        try:
            self.logger.info(
                "Starting up the testing cluster using the provisioner")
            self.provisioner.apply()
        except Exception as e:
            raise Exception("Provisioner failed to apply()") from e

        self.func(self.config, self.provisioner)

    def run2(self):
        """ get the provisioner but start the provisioner before returning

        This is preferable to the raw provisioner in cases where you want a running
        cluster so that the cluster startup cost does not get reflected in the
        first test case which uses the fixture.  Also it can tear itself down

        You can still use provisioner.apply() update the resources if the provisioner
        can handle it.
        """
        self.logger.info("Running metta provisioner up()", id)

        conf = self.config.load("config")
        provisioner_config = self.config.load("provisioner")

        try:
            self.logger.info(
                "Preparing the testing cluster using the provisioner")
            self.provisioner.prepare()
        except Exception as e:
            raise Exception("Provisioner failed to prepare()") from e
        try:
            self.logger.info(
                "Starting up the testing cluster using the provisioner")
            self.provisioner.apply()
        except Exception as e:
            raise Exception("Provisioner failed to apply()") from e

        self.func(self.config, self.provisioner)

        if conf.get("options.destroy-on-finish", default=False):
            try:
                self.logger.info(
                    "Stopping the test cluster using the provisioner as directed by config")
                self.provisioner.destroy()
            except Exception as e:
                raise Exception("Provisioner failed to destroy()") from e
        else:
            self.logger.info(
                "Leaving test infrastructure in place on shutdown")




import re
import logging

from configerus.config import Config

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta_common.dict_output import DictOutputPlugin
from mirantis.testing.metta_common.text_output import TextOutputPlugin

OUTPUT_FORMAT_MATCH_PATTERN = r'(?P<output>(\w+)+)(\/(?P<base>[\-\.\w]+))?'
""" A regex pattern to identify outputs that should be embedded """

logger = logging.getLogger('configerus.contrib.files:output')


class ConfigFormatOutputPlugin:
    """   """

    def __init__(self, config: Config, instance_id: str):
        """  """
        self.config = config
        self.instance_id = instance_id

        self.pattern = re.compile(OUTPUT_FORMAT_MATCH_PATTERN)
        """ Regex patter for identifying an output replacement """

        self.environment = None
        """ environment which contains the outputs. Must be added """

    def set_environemnt(self, environment: Environment):
        """ set the output environment

        @NOTE this is obligatory

        """
        self.environment = environment

    def format(self, key, default_label: str):
        """ Format a string by substituting config values

        Parameters
        ----------

        key: a string that should gies instructions to the formatter on how to
            create a format replacementsrmed

        default_source : if format/replace patterns don't have a source defined
            then this is used as a source.

        """
        # if the entire key is the match, then replace whatever type we get
        # out of the config .get() call
        match = self.pattern.fullmatch(key)
        if not match:
            raise KeyError(
                "Could not interpret Format action key '{}'".format(key))

        output = match.group('output')

        try:
            output = self.environment.fixtures.get_plugin(
                type=Type.OUTPUT, instance_id=output)

            if isinstance(output, DictOutputPlugin):
                base = match.group('base')
                if base is not None:
                    return output.get_output(base)
                return output.get_output()
            elif isinstance(output, TextOutputPlugin):
                return output.get_output()
            elif hasattr(output, 'get_output'):
                return output.get_output()

        except KeyError as e:
            raise KeyError(
                "Config replace for output failed as output '{}' was not found, and no default value was suggested.".format(output)) from e



"""

A helper plugin for downloading executables from the internet to help other
Plugins which might rely on external executables .

@NOTE remote urls may be tarballs/zips

@NOTE it goes without saying that downloading executables comes with an inherit
risk.  The request validates any certs, but it is up to you to make sure that
no config is in place which downloads dangerous bins.

The idea here is that you can tell your system to put bins from a list of
urls into a path, and have that path included in PATH in your python environment.
If you want, use a common system path like '/usr/local/bin' or use a temporary
or local path if you want to isolate versions for your test suite.

There are two ways you can use this:

1. put a list of bins into the plugin config, and it will download when it is
constructed.  This de-couples acquiring bins from plugins that may use them, and
does all of the downloading up front before all fixtures are created.
If another fixture needs a bin in its own constructor, just make sure that this
fixture is created first.

2. any other code with access to the environment can access the binhelper plugin
and provide details to the `get_bin` method.  This couples the plugins, and
means that you will have to ensure that this plugin is available and handle
dependencies yourself.

"""
from typing import Dict
import shutil
import os
import logging
import requests
import tarfile
import zipfile
import re
import stat
import platform

from configerus.loaded import LOADED_KEY_ROOT
from configerus.contrib.jsonschema.validate import PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL
from configerus.validator import ValidationError

from mirantis.testing.metta.plugin import METTAPlugin
from mirantis.testing.metta.environment import Environment

logger = logging.getLogger('metta.common.utility:binhelper')

METTA_PLUGIN_ID_UTILITY_BINHELPER = 'bin-helper'
""" utility plugin_id for the bin-helper plugin """

BINHELPER_UTILITY_CONFIG_LABEL = 'binhelper'
""" Config label used by default to load binhelper configuration """

BINHELPER_UTILITY_CONFIG_BASE_LOCALPATH = 'path.local'
""" config base for what the local bin path should be """
BINHELPER_UTILITY_CONFIG_BASE_ADDTOPATH = 'path.add_to_path'
""" config base for whether or not we need to modify the env PATH and add the path """
BINHELPER_UTILITY_CONFIG_BASE_PLATFORMS = 'platforms'
""" config base for list of platform bins to load on construction """

BINHELPER_UTILITY_CONFIG_BASE_BIN_URL = 'url'
""" config base inside bin for bin url for downloading """
BINHELPER_UTILITY_CONFIG_BASE_BIN_VERSION = 'version'
""" config base inside bin for bin version """
BINHELPER_UTILITY_CONFIG_BASE_BIN_COPYPATHS = 'copy'
""" config base inside bin for a map of bins name->path-inpackage """

BINHELPER_CONFIG_JSONSCHEMA = {
    'type': 'object',
    'path': {
        'type': 'object',
        'properties': {
            'local': {'type': 'string'},
            'add_to_environ': {'type': 'bool'}
        },
        'required': ['path']
    },
    'platforms': {
        '$ref': '#/definitions/platform'
    },
    'definitions': {
        'bin': {
            'type': 'object',
            'properties': {
                'url': {'type': 'string'},
                'version': {'type': 'string'},
                'copy': {'type': 'object'}
            },
            'required': ['url']
        },
        'platform': {
            '$ref': '#/definitions/bin'
        },
    },
    'required': ['path', 'platforms']
}
""" JSONSCHEMA for the binhelper config """
BINHELPER_CONFIG_VALIDATE_TARGET = {
    PLUGIN_ID_VALIDATE_JSONSCHEMA_SCHEMA_CONFIG_LABEL: BINHELPER_CONFIG_JSONSCHEMA}
""" configerus validate target for binhelper config """


class DownloadableExecutableUtility(METTAPlugin):
    """ A helper to help you handle those pesky executables for tools which you need to download """

    def __init__(self, environment: Environment, instance_id: str,
                 label: str = BINHELPER_UTILITY_CONFIG_LABEL, base: str = LOADED_KEY_ROOT):
        """

        Parameters:
        -----------

        """
        METTAPlugin.__init__(self, environment, instance_id)

        loaded = self.environment.config.load(label)
        try:
            loaded.get(
                base,
                validator=BINHELPER_CONFIG_VALIDATE_TARGET)
        except KeyError as e:
            raise ValueError("Bin-Helper configuration is missing") from e
        except ValidationError as e:
            raise ValueError(
                "Bin-Helper received invalid configuration: {}".format(e)) from e

        self.local_path = loaded.get(
            [base, BINHELPER_UTILITY_CONFIG_BASE_LOCALPATH])

        add_to_path = loaded.get([base,
                                  BINHELPER_UTILITY_CONFIG_BASE_ADDTOPATH])
        if add_to_path:
            os.environ["PATH"] += os.pathsep + \
                os.path.realpath(self.local_path)

        platforms = loaded.get([base,
                                BINHELPER_UTILITY_CONFIG_BASE_PLATFORMS])
        current_platform = '{}-{}'.format(platform.system(),
                                          platform.machine())

        if current_platform not in platforms:
            logger.warn("BinHelper doesn't have configuration for your platform '{}', so it won't download any bins.".format(
                current_platform))
            return

        bins = platforms[current_platform]
        for bin_id in list(bins.keys()):
            url = loaded.get([base,
                              BINHELPER_UTILITY_CONFIG_BASE_PLATFORMS,
                              current_platform,
                              bin_id,
                              BINHELPER_UTILITY_CONFIG_BASE_BIN_URL])
            version = loaded.get([base,
                                  BINHELPER_UTILITY_CONFIG_BASE_PLATFORMS,
                                  current_platform,
                                  bin_id,
                                  BINHELPER_UTILITY_CONFIG_BASE_BIN_VERSION])
            copypaths = loaded.get([base,
                                    BINHELPER_UTILITY_CONFIG_BASE_PLATFORMS,
                                    current_platform,
                                    bin_id,
                                    BINHELPER_UTILITY_CONFIG_BASE_BIN_COPYPATHS])
            self.get_bin(
                name=bin_id,
                url=url,
                version=version,
                copypaths=copypaths)

    def get_bin(self, name: str, url: str, version: str,
                copypaths: Dict[str, str] = None):
        """ get a remote bin package, and put any bin contents into a bin path

        Make sure that we have the bin in scope; download it if we don't

        """
        path = shutil.which(name, os.X_OK)

        if path is None:
            logger.info(
                "bin-helper can't find {}. Downloading it from {}".format(name, url))

            # First get the url
            r = requests.get(url, allow_redirects=True)
            r.raise_for_status()

            # try to decide on a name for the download
            file_name = None
            content_disposition = r.headers.get('content-disposition')
            if content_disposition is not None:
                file_name_matches = re.findall(
                    'filename=(.+)', content_disposition)
                if file_name_matches:
                    file_name = file_name_matches[0]
            if file_name is None:
                # @NOTE this may not be a reliable way of getting a name from a url
                file_name = os.path.basename(url)

            # write the file into our local path
            local_file = os.path.join(self.local_path, file_name)
            with open(local_file, 'wb') as f:
                f.write(r.content)

            # downloaded url is a tarfile.  Copy only the files out of the tarfile
            # that are suggested by the "copy" part of the config, and make sure
            # to 'chmod a+x' them.
            if tarfile.is_tarfile(local_file):
                with tarfile.open(local_file) as tf:
                    members = tf.getmembers()
                    for (bin_name, from_name) in copypaths.items():
                        for member in members:
                            if member.name == from_name:
                                member.name = bin_name
                                tf.extract(member, self.local_path)
                                bin_path = os.path.join(
                                    self.local_path, bin_name)
                                os.chmod(
                                    bin_path, os.stat(bin_path).st_mode | stat.S_IEXEC)
                                break
                os.remove(local_file)

            # downloaded url is a zipfile.  Copy only the files out of the zipfile
            # that are suggested by the "copy" part of the config, and make sure
            # to 'chmod a+x' them.
            elif zipfile.is_zipfile(local_file):
                with zipfile.ZipFile(local_file) as zf:
                    members = zf.infolist()
                    for (bin_name, from_name) in copypaths.items():
                        for member in members:
                            if member.filename == from_name:
                                bin_path = os.path.join(
                                    self.local_path, bin_name)
                                source = zf.open(member)
                                target = open(bin_path, "wb")
                                with source, target:
                                    shutil.copyfileobj(source, target)
                                os.chmod(
                                    bin_path, os.stat(bin_path).st_mode | stat.S_IEXEC)
                os.remove(local_file)

            # this case doesn't really make sense, but is possible based on config
            # where you specified a download url and multiple copy paths.
            # I don't think it has any actual usecases, but I wrote it anyway.
            elif copypaths:
                for (bin_name, from_name) in copypaths.items():
                    bin_path = os.path.join(self.local_path, bin_name)
                    from_path = os.path.join(self.local_path, from_name)
                    os.rename(from_path, bin_path)
                    os.chmod(
                        bin_path, os.stat(bin_path).st_mode | stat.S_IEXEC)

            # In this case, you specified no `copy` items, so your URL must be
            # a single file that we downloaded.  Let's just make sure it is
            # executable.
            elif os.path.isfile(local_file):
                os.chmod(
                    local_file,
                    os.stat(local_file).st_mode | stat.S_IEXEC)



import logging
from typing import Dict, Any
import time
import json
import yaml

from mirantis.testing.metta.plugin import Type
from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.cli import CliBase

from .sonobuoy import METTA_PLUGIN_ID_SONOBUOY_WORKLOAD, Status

logger = logging.getLogger('metta.cli.sonobuoy')

METTA_PLUGIN_ID_SONOBUOY_cli = 'metta_sonobuoy_cli'
""" cli plugin_id for the sonobuoy plugin """


class SonobuoyCliPlugin(CliBase):

    def fire(self):
        """ return a dict of commands for sonobuoy workloads if one hase been registered."""
        if self.environment.fixtures.get_fixture(
                type=Type.WORKLOAD, plugin_id=METTA_PLUGIN_ID_SONOBUOY_WORKLOAD, exception_if_missing=False) is not None:
            return {
                'contrib': {
                    'sonobuoy': SonobuoyGroup(self.environment)
                }
            }
        else:
            return {}


class SonobuoyGroup():
    """ Sonobuoy workload plugin cli command group """

    def __init__(self, environment: Environment):
        self.environment = environment

    def _select_fixture(self, instance_id: str = ''):
        """ Pick a matching workload plugin """
        try:
            if instance_id:
                return self.environment.fixtures.get_fixture(
                    type=Type.WORKLOAD, plugin_id=METTA_PLUGIN_ID_SONOBUOY_WORKLOAD, instance_id=instance_id)
            else:
                # Get the highest priority provisioner
                return self.environment.fixtures.get_fixture(
                    type=Type.WORKLOAD, plugin_id=METTA_PLUGIN_ID_SONOBUOY_WORKLOAD)

        except KeyError as e:
            raise ValueError(
                "No usable kubernetes client was found for sonobuoy to pull a kubeconfig from: {}".format(e)) from e

    def _select_instance(self, instance_id: str = ''):
        """ create a sonobuoy workload plugin instance """
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        # @TODO allow filtering of kubernetes client instances
        instance = plugin.create_instance(self.environment.fixtures)
        return instance

    def info(self, instance_id: str = '', deep: bool = False):
        """ get info about a provisioner plugin """
        fixture = self._select_fixture(instance_id=instance_id)

        info = {
            'fixture': {
                'type': fixture.type.value,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                info.update(fixture.plugin.info(True))

        return json.dumps(info, indent=2)

    def status(self, instance_id: str = ''):
        """ get active sonobuoy status """
        instance = self._select_instance(instance_id=instance_id)
        status = instance.status()

        status_info = {
            'status': status.status.value,
            'plugins': {plugin_id: status.plugin(plugin_id) for plugin_id in status.plugin_list()}
        }

        return json.dumps(status_info, indent=2)

    def crb(self, instance_id: str = '', remove: bool = False):
        """ create the crb needed to run sonobuoy """
        instance = self._select_instance(instance_id=instance_id)

        if remove:
            instance._delete_k8s_crb()
        else:
            instance._create_k8s_crb()

    def run(self, instance_id: str = '', wait: bool = False):
        """ run sonobuoy workload """
        instance = self._select_instance(instance_id=instance_id)
        instance.run(wait=wait)

    def wait(self, instance_id: str = '', step: int = 5, limit: int = 1000):
        """ wait until no longer running """
        instance = self._select_instance(instance_id=instance_id)
        print('{')
        for i in range(0, limit):
            status = instance.status()
            status_info = {
                'status': status.status.value,
                'plugins': {plugin_id: status.plugin(plugin_id)['status'] for plugin_id in status.plugin_list()}
            }

            print("{}: {},".format(i, status_info))
            if status.status not in [Status.RUNNING]:
                break

            time.sleep(step)
        print('}')

    def destroy(self, instance_id: str = '', wait: bool = False):
        """ remove all sonobuoy infrastructure """
        instance = self._select_instance(instance_id=instance_id)
        instance.destroy(wait=wait)

    def logs(self, instance_id: str = '', follow: bool = False):
        """ sonobuoy logs """
        instance = self._select_instance(instance_id=instance_id)
        instance.logs(follow=follow)

    def retrieve(self, instance_id: str = ''):
        """ retrieve the results from the sonobuoy workload instance """
        instance = self._select_instance(instance_id=instance_id)
        try:
            instance.retrieve()
        except Exception as e:
            logger.error("Retrieve failed: {}".format(e))
